/*
 * SwitchDesk module for VirtuaWin
 * Created by Alasdair McCaig
 * Alasdair.McCaig@Poptel.net
 *
 * Changes in Version 2.0 (November 2015) by Ben Moore
 * ======================
 * Made the code more portable - relative paths are now stored for cached
 *   wallpaper images.
 * Stability improvements.  
 *   NOTE: Cached filename format has been changed
 *       from:  switchdesk.<deskid>.<ext>
 *         to:  switchDesk-<deskid>.<ext>
 *   and config filename has been changed
 *       from:  switchdesk.<deskid>
 *         to:  switchDesk-<deskid>.cfg
 *   to avoid some potential issues.  Rename your existing files in your user
 *   directory as necessary if you want to keep your current settings.
 * In-memory desktop switch comparisons - After an initial load of a desktop's
 *   settings from config, further operations are handled in memory.  Writes
 *   to config files or registry entries are only done when an actual change
 *   needs to be made.  The less those expensive write operations need to be 
 *   done, the better.
 * Made the code more readable and easier to understand.
 *
 * Changes in Version 1.8 (October 2011) by Ethan Warth (a.k.a. redyoshi49q)
 * ======================
 * Set to load desktop settings on module startup, and restore them on module exit
 * This should make the module more friendly for the host computer when
 *   using VirtuaWin Portable
 *
 * Changes in Version 1.7 (June 2007) by Brian Vanderburg II & Ben Moore 
 * ======================
 * Fixed memory leak
 * Fixed to use VW_USERAPPPATH path for desktop settings files
 * Correctly sets/restored centered/stretched status
 * Copy wallpaper to switchdesk.<deskid>.<ext>, fixes jpg to 'Wallpaper1.bmp' issue
 *
 * Changes in Version 1.6 (December 2005 by Johan Piculell)
 * ======================
 * Updated for VirtuaWin 2.11
 *
 * Changes in Version 1.5 (July 2000)
 * ======================
 * Multi-User Support
 *
 * Version 1.4 (June 2000)
 * Version 1.3 (February 2000)
 * Version 1.0 Created by Alasdair McCaig 1999
 *
 * Bug Fixes
 * =========
 * 1) A memory leak has been fixed
 * 2) Registry permissions problem fixed. (Should fix prob for Win2000 as well)
 *
 * Both of these fixes where made possible by Timo Wilson... Cheers
 *
 *
 * Platforms
 * =========
 * Windows 95, 98, NT 4 sp6, 2000, XP
 *
 */

// Standard Includes
#include <windows.h>
#include <string.h>
#include <stdio.h>
#include <tchar.h>
#include "Messages.h"
#include "SwitchDesk.h"

#define MAX_DESK 32

// Define a struct to hold wallpaper settings.
typedef struct {
	char wallpaper[MAX_PATH];
	char tile[2];
	char style[2];
} desktop_info;

// Declare and initialize global variables .
desktop_info desktop[MAX_DESK + 1] = { };	// Array of wallpaper settings for all of our desktops.
int initialized=0;
HWND vwHandle;     							// Handle to VirtuaWin
static int deskid=-1;
static char userConfigPath[MAX_PATH] = { 0 };
static char switchDeskFileNameBase[20] = "switchDesk-";		// Base name of config and cache files.

// ---------------------------------------------------------------------------------------------------------------------
// Callback message handler.
//
LRESULT CALLBACK MainWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam) 
{
    COPYDATASTRUCT* cds;
    
    switch (msg)
    {
    case MOD_INIT: // This must be taken care of in order to get the handle to VirtuaWin. 
        // The handle to VirtuaWin comes in the wParam .
        vwHandle = (HWND) wParam; // Should be some error handling here if NULL .
		
        if(!initialized)
        {
			// Initialize tile to null so we can use it as an indicator that data has been loaded.
			int i;
			for(i=0; i<MAX_DESK + 2; i++) {
					desktop[i].tile[0] = '\0';
			}
			
            // Request the VW user's path (will be handled and returned by a separate WM_COPYDATA callback).
            SendMessage(vwHandle, VW_USERAPPPATH, 0, 0);
            if(!initialized)
            {
                MessageBox(hwnd,"VirtuaWin failed to send the UserApp path.","SwitchDesk Error", MB_ICONWARNING);
                exit(1) ;
            }
        }
		saveDesk(0);  // Save the current (non-VirtuaWin) desktop as desktop 0 so that it can be restored on exit.
        switchDesk(SendMessage(vwHandle,VW_CURDESK,0,0)) ;	// Switch to our first VirtuaWin desktop.
        break;
    
    case WM_COPYDATA:
        // Make sure this is the correct "VW_USERAPPPATH" message we are looking to handle here.
        cds = (COPYDATASTRUCT*)lParam;
        if(cds->dwData == 0 - VW_USERAPPPATH)
        {
            initialized = 1 ;
            // It seems that this gets called several times.
            // No problem since userConfigPath is static array and not
            //   dynamically allocated.
            strncpy(userConfigPath, (LPSTR)(cds->lpData), MAX_PATH - 1);
            return TRUE; // We should return true if we process this message 
        }
        break;
    
    case MOD_QUIT: // This must be handled, otherwise VirtuaWin can't shut down the module .
		switchDesk(0); // Switch to original desktop before module closes.
        PostQuitMessage(0);
        break;
    case MOD_SETUP: 
        MessageBox(vwHandle, "Created by Alasdair McCaig 2000 \n\r\n\r2.0 Enhancements by Ben Moore \n\r\n\rUse the Display Control Panel to change the background\n\rfor each desktop.", "SwitchDesk 2.0", 0);
        break;
    
    case MOD_CHANGEDESK:
    case MOD_STEPLEFT:
    case MOD_STEPRIGHT:
    case MOD_STEPUP:
    case MOD_STEPDOWN:
        switchDesk(lParam); // Switch to the new Desktop.
        break;
    
    case WM_DESTROY:
		switchDesk(0); // Switch to the original desktop before exit.
        PostQuitMessage(0);
        break;
    default:
        return DefWindowProc(hwnd, msg, wParam, lParam);
    }
    
    return 0;
}

// ---------------------------------------------------------------------------------------------------------------------
// Main module startup function.
//
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, INT nCmdShow)
{
    WNDCLASS wc;
    MSG msg;
    
    memset(&wc, 0, sizeof(WNDCLASS));
    wc.style = 0;
    wc.lpfnWndProc = (WNDPROC)MainWndProc;
    wc.hInstance = hInstance;
    wc.lpszClassName = "SwitchDesk.exe";
    
    if (!RegisterClass(&wc)) return 0;
    
    // In this example, the window is never shown.
    if (CreateWindow("SwitchDesk.exe", 
                     "SwitchDesk", 
                     WS_POPUP,
                     CW_USEDEFAULT, 
                     0, 
                     CW_USEDEFAULT, 
                     0,
                     NULL,
                     NULL,
                     hInstance,
                     NULL) == (HWND)0)
        return 0;
    
    // main message loop
    while (GetMessage(&msg, NULL, 0, 0)) {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }
    return msg.wParam;
}



// ---------------------------------------------------------------------------------------------------------------------
// Save desktop.
//
void saveDesk(int deskid)
{
    HKEY hkey=NULL;
    DWORD dwType;
    long lResult;
	DWORD cbData = MAX_PATH;
    char configFile[MAX_PATH];
    char wallpaperCachedFile[MAX_PATH];
    char *ext;
    FILE* fp;
	desktop_info deskFromRegistry = { };
    
    if(deskid < 0)
        return ;
    
	// Store the fully qualified location of our config file for this desktop into configFile.
    sprintf(configFile, "%s%s%i%s", userConfigPath, switchDeskFileNameBase, deskid, ".cfg");
    
    // Get desktop info from registry
	if(RegOpenKeyEx(HKEY_CURRENT_USER, "Control Panel\\Desktop", 0,KEY_READ, &hkey) == 0)
	{
		// Get wallpaper
		lResult = RegQueryValueEx(hkey, "Wallpaper", NULL, &dwType, (LPBYTE)deskFromRegistry.wallpaper, &cbData);
		// Add null termination to string on possibility that value in registry was not properly terminated.
		deskFromRegistry.wallpaper[cbData] = 0;
		
		// Get Tiled
		lResult = RegQueryValueEx(hkey, "TileWallpaper", NULL, &dwType, (LPBYTE)deskFromRegistry.tile, &cbData);
		// Add null termination to string on possibility that value in registry was not properly terminated.
		deskFromRegistry.tile[cbData] = 0;
		
		// Get Style
		lResult = RegQueryValueEx(hkey, "WallpaperStyle", NULL, &dwType, (LPBYTE)deskFromRegistry.style, &cbData);
		// Add null termination to string on possibility that value in registry was not properly terminated.
		deskFromRegistry.style[cbData] = 0;
		
		CloseHandle(hkey);

		
		// Compare to current in-memory values of desktop and save if registry has any new differences.
		if(	    strcmpi(deskFromRegistry.wallpaper, desktop[deskid].wallpaper) != 0
			||  strcmpi(deskFromRegistry.tile, desktop[deskid].tile) != 0
			||  strcmpi(deskFromRegistry.style, desktop[deskid].style) != 0
		) {
	
			// Open file to save data.
			if((fp = fopen(configFile, "w")) != NULL)
			{		
				if (deskid != 0) {
					// Generate filename used to cache wallpaper for this desktop.
					// Do some extension handling due to how older windows versions used to convert all wallpapers to bitmaps.
					ext = strrchr(deskFromRegistry.wallpaper, '.');
					if(ext == NULL)
						ext = ".bmp";
					sprintf(wallpaperCachedFile, "%s%s%i%s", userConfigPath, switchDeskFileNameBase, deskid, ext);
				
					// Only copy wallpaper to a cache file if the wallpaper set in registry is not already the cached file.
					if(strcmpi(wallpaperCachedFile, deskFromRegistry.wallpaper) != 0)
						CopyFile(deskFromRegistry.wallpaper, wallpaperCachedFile, FALSE);
					
					// Save name of cache file to config.
					fprintf(fp, "%s\n", strrchr(wallpaperCachedFile, '\\') + 1);
					sprintf(desktop[deskid].wallpaper, "%s", wallpaperCachedFile);
				} 
				else { 
					// for deskid == 0, simply save the original file's name.
					fprintf(fp, "%s\n", deskFromRegistry.wallpaper);
					sprintf(desktop[deskid].wallpaper, "%s", deskFromRegistry.wallpaper);
				}
				
				// Save Tiled
				fprintf(fp, "%s\n", deskFromRegistry.tile);  
				sprintf(desktop[deskid].tile, "%s", deskFromRegistry.tile);
				
				// Save Style
				fprintf(fp, "%s\n", deskFromRegistry.style);  
				sprintf(desktop[deskid].style, "%s", deskFromRegistry.style);
				
				fclose(fp);
			}
		}
        
    }
}



// ---------------------------------------------------------------------------------------------------------------------
// Load desktop.
//
void loadDesk(int deskid)
{
    HKEY hkey=NULL;
	DWORD dwType;
    long lResult;
	DWORD cbData = MAX_PATH;
    char tmpstr[MAX_PATH];
    char configFile[MAX_PATH];
    FILE* fp;
	desktop_info deskFromRegistry = { };
    
	// Store the fully qualified location of our config file for this desktop into configFile.
    sprintf(configFile, "%s%s%i%s", userConfigPath, switchDeskFileNameBase, deskid, ".cfg");

	// If settings have not yet been loaded into memory for this desktop, load settings from config file.
	if (desktop[deskid].tile[0] == '\0') {
		if((fp = fopen(configFile, "r")) != NULL) {
			// Get wallpaper image filename from first line of config file.
			if (fgets(tmpstr, 4098, fp) != NULL) {
				// If there is no path information, prepend the userConfigPath before saving to registry.
				if (strrchr(tmpstr, '\\') == NULL ) 
					sprintf(desktop[deskid].wallpaper, "%s%s", userConfigPath, tmpstr);
				else
					sprintf(desktop[deskid].wallpaper, "%s", tmpstr);
				// Replace newline at end of string from file with a null to later store filename properly in registry.
				desktop[deskid].wallpaper[strcspn(desktop[deskid].wallpaper, "\n")] = 0;
			}
			// Get wallpaper tile setting from second line of config file.
			if (fgets(tmpstr, 4098, fp) != NULL) {
				sprintf(desktop[deskid].tile, "%s", tmpstr);
				// Replace newline at end of string from file with a null to later store filename properly in registry.
				desktop[deskid].tile[strcspn(desktop[deskid].tile, "\n")] = 0;
			}
			// Get wallpaper style setting from third line of config file.
			if (fgets(tmpstr, 4098, fp) != NULL) {
				sprintf(desktop[deskid].style, "%s", tmpstr);
				// Replace newline at end of string from file with a null to later store filename properly in registry.
				desktop[deskid].style[strcspn(desktop[deskid].style, "\n")] = 0;
			}
			
			fclose(fp);
		}
	}
	
	// Get current desktop info from registry
	if(RegOpenKeyEx(HKEY_CURRENT_USER, "Control Panel\\Desktop", 0,KEY_READ, &hkey) == 0)
	{
		// Get wallpaper
		lResult = RegQueryValueEx(hkey, "Wallpaper", NULL, &dwType, (LPBYTE)deskFromRegistry.wallpaper, &cbData);
		// Add null termination to string on possibility that value in registry was not properly terminated.
		deskFromRegistry.wallpaper[cbData] = 0;
		
		// Get Tiled
		lResult = RegQueryValueEx(hkey, "TileWallpaper", NULL, &dwType, (LPBYTE)deskFromRegistry.tile, &cbData);
		// Add null termination to string on possibility that value in registry was not properly terminated.
		deskFromRegistry.tile[cbData] = 0;
		
		// Get Style
		lResult = RegQueryValueEx(hkey, "WallpaperStyle", NULL, &dwType, (LPBYTE)deskFromRegistry.style, &cbData);
		// Add null termination to string on possibility that value in registry was not properly terminated.
		deskFromRegistry.style[cbData] = 0;
		
		CloseHandle(hkey);
		
		// Open registry for writing and if values are different, write new values to registry as necessary.
		if(RegOpenKeyEx(HKEY_CURRENT_USER, "Control Panel\\Desktop", 0,KEY_SET_VALUE, &hkey) == 0)
		{
			if (strcmpi(deskFromRegistry.wallpaper, desktop[deskid].wallpaper) != 0)
				RegSetValueEx(hkey, "Wallpaper", 0, REG_SZ, desktop[deskid].wallpaper, (strlen(desktop[deskid].wallpaper)*sizeof(char)));
			if (strcmpi(deskFromRegistry.tile, desktop[deskid].tile) != 0)
				RegSetValueEx(hkey, "TileWallpaper", 0, REG_SZ, desktop[deskid].tile, (strlen(desktop[deskid].tile)*sizeof(char)));
			if (strcmpi(deskFromRegistry.style, desktop[deskid].style) != 0)
				RegSetValueEx(hkey, "WallpaperStyle", 0, REG_SZ, desktop[deskid].style, (strlen(desktop[deskid].style)*sizeof(char)));
			
			CloseHandle(hkey);
		}
    }
}



// ---------------------------------------------------------------------------------------------------------------------
// Switch from one desktop to another.
//
void switchDesk(int newdesk)
{
    if ( (newdesk >= 0) && (newdesk < MAX_DESK) && (newdesk != deskid)) {
		// Save the settings of the old desktop.
        saveDesk(deskid);
        // Switch to the new desktop.
        deskid=newdesk;
        loadDesk(deskid);
        SystemParametersInfo(SPI_SETDESKWALLPAPER, 0, NULL, SPIF_SENDCHANGE );
    }    
}

